#Include Once "{TKEY}Support.inc"
#Include Once "SharedSupport.inc"

' Unique collection of {Tkey}'s.
'
' Property Count() As Long - Get/Set the # of data items.  Setting alters the capacity.  Shrinking will cause a clear.
' Property Contains( pKey ) As Long - Return -1 if key is in collection.  0 if key is not in collection.
' Property Index( pKey As {TKEY} ) As Long - Get Index of key or -1 if key is not in collection.
' Property Item( pKey ) As {TKEY} - Get key value by key.
' Property Key( index ) As {TKEY}  - Get key value by index.
'
' Method Add( pKey ) As Long - Add key to collection.  Return index of new or existing key.
' Method Clear() - Clear the collection
'
' Property HashCount() As Long - Get/Set capacity of hash table.  Setting will clear collection.
' Property ResizeAmount() As Long - Get/Set the # of items to grow data collection by.

Sub New{TKEY}Bag( obj As i{TKEY}Bag, ByVal hashSize As Long, Optional resizeAmount As Long )
	' Passing the object to create by reference is 1) slightly faster than using a function, and 2) results in correct ref counts (functions leave an extra ref count in 10.03)
	obj = Class "c{TKEY}Bag"
	obj.HashCount = hashSize
	If Not IsMissing( resizeAmount ) Then
		obj.ResizeAmount = resizeAmount
	End If
End Sub

Class c{TKEY}Bag
	Instance hash_() As Long
	Instance next_() As Long
	Instance key_() As {TKEY}
	Instance hashCount_ As Long
	Instance dataCount_ As Long
	Instance ubData_ As Long
	Instance dataResizeAmount_ As Long

	Class Method Create()
		dataResizeAmount_ = 100
	End Method

	Interface i{TKEY}Bag
		Inherit IUnknown

		Property Get HashCount() As Long
			Property = hashCount_
		End Property
		Property Set HashCount( value As Long )
			Me.ResizeHash( value )
		End Property

		Property Get ResizeAmount() As Long
			Property = dataResizeAmount_
		End Property
		Property Set ResizeAmount( ByVal value As Long )
			If value>0 Then dataResizeAmount_ = value
		End Property

		Property Get Count() As Long
			Property = dataCount_
		End Property
		Property Set Count( ByVal value As Long )
			If value> dataCount_ Then Me.ResizeData( value )
		End Property

		Method Clear()
			Me.ResizeHash( 0 )
			Me.ResizeData( 0 )
		End Method

		' Returns index of added item
		Method Add( {KEYREF} pKey As {TKEY} ) As Long
			#Debug Code Off
			Register index As Long
			Register lastIndex As Long
			Local hc As Long

			hc = {HASH}( pKey, hashCount_ )

			index = hash_( hc )

			If index = -1 Then
				' First entry
				If dataCount_>=ubData_ Then Me.ResizeData( ubData_+dataResizeAmount_ )
				hash_( hc ) = dataCount_
			Else
				' Collision
				Do
					If {CMP}( pKey, key_( index ) ) Then
						lastIndex = index
						index = next_( index )
					Else
						Method = index
						Exit Method
					End If
				Loop While index>-1
				If dataCount_>=ubData_ Then Me.ResizeData( ubData_+dataResizeAmount_ )
				next_( lastIndex ) = dataCount_
			End If
			next_( dataCount_ ) = -1
			key_( dataCount_ ) = pKey
			Method = dataCount_
			Incr dataCount_
			#Debug Code On
		End Method

		' -1 if found, 0 if not found
		Property Get Contains( {KEYREF} pKey As {TKEY} ) As Long
			Property = ( Me.Index( pKey ) > -1 )
		End Property

		Property Get Index( pKey As {TKEY} ) As Long
			' Returns index of key
			' Returns -1 if not found
			' Returns 0 - dataCount_-1 if found
			Register index As Long
			Local hc As Long

			hc = {HASH}( pKey, hashCount_ )
			index = hash_( hc )

			While index > -1
				If {CMP}( pKey, key_( index ) ) Then
					index = next_( index )
				Else
					Property = index
					Exit Property
				End If
			Wend
			Property = -1
		End Property

		Property Get Key( ByVal index As Long ) As {TKEY}
			If index>=0 And index< dataCount_ Then
				Property = key_( index )
			End If
		End Property

		Property Get Item( {KEYREF} pKey As {TKEY} ) As {TKEY}
			Property = Me.Key( Me.Index( pKey ) )
		End Property
	End Interface

	Class Method ResizeData( ByVal value As Long )
		If value>0 Then
			If value < dataCount_ Then
				' Have to clear hash if shrinking, otherwise there will be invalid entries
				If hashCount_>0 Then Memory Fill VarPtr( hash_(0) ), value*SizeOf(value), Byte &HFF
				dataCount_ = 0
			End If
			ReDim Preserve next_( value-1 )
			ReDim Preserve key_( value-1 )
			ubData_ = value
		Else
			Erase next_()
			Erase key_()
			ubData_ = 0
			dataCount_ = 0
			' Have to clear hash if shrinking, otherwise there will be invalid entries
			If hashCount_>0 Then Memory Fill VarPtr( hash_(0) ), value*SizeOf(value), Byte &HFF
		End If
	End Method

	Class Method ResizeHash( ByVal value As Long )
		' Clears bag
		value = CollectionSize( value ): ' Make a prime >= value
		ReDim hash_( value-1 )
		Memory Fill VarPtr( hash_(0) ), value*SizeOf(value), Byte &HFF
		dataCount_ = 0
		hashCount_ = value
	End Method
End Class
