#Include Once "{TKEY}Support.inc"
#Include Once "SharedSupport.inc"

' Property Count() As Long - Get # of items in collection, Set capacity of items.  Shrinking will cause collection to be cleared.
' Property Contains( pKey ) As Long - Retuns -1 if key is in collection, 0 if key is not in collection.
' Property Index( pKey ) As Long - Get index of key or -1 if key not in collection.
' Property Item( pKey ) As {TData} - Get/Set data value by key
' Property Key( index ) As {TKey} - Get key value at an index
' Property Value( index ) As {TData} - Get/Set data value by index
'
' Method Add( pKey, value ) As Long - Add an item, return index of new/existing item
' Method Clear() - Clear collection
'
' Property HashCount As Long - Get/Set size of hash table. Setting will clear collection
' Property ResizeAmount() As Long - Get/Set amount to resize data collection

Sub New{TKEY}To{TDATA}Map( obj As i{TKEY}To{TDATA}Map, ByVal hashSize As Long, Optional resizeAmount As Long )
	' Passing the object to create by reference is 1) slightly faster than using a function, and 2) results in correct ref counts (functions leave an extra ref count in 10.03)
	obj = Nothing
	obj = Class "c{TKEY}To{TDATA}Map"
	obj.HashCount = hashSize
	If Not IsMissing( resizeAmount ) Then
		obj.ResizeAmount = resizeAmount
	End If
End Sub

Sub Dup{TKEY}To{TDATA}Map( obj As i{TKEY}To{TDATA}Map, src As i{TKEY}To{TDATA}Map )
	Local i As Long
	Local value As {TKEY}
	Local sf As {TDATA}
	obj = Nothing
	If IsFalse IsInterface( src, i{TKEY}To{TDATA}Map ) Then Exit Sub
	obj = Class "c{TKEY}To{TDATA}Map"
	obj.HashCount = src.HashCount
	obj.ResizeAmount = src.ResizeAmount
	For i=0 To src.Count-1
		sf = src.Value(i)
		value = src.Key( i )
		obj.Add( value, sf )
		Reset sf
	Next
End Sub

Class c{TKEY}To{TDATA}Map
	Instance hash_() As Long
	Instance next_() As Long
	Instance key_() As {TKEY}
	Instance value_() As {TDATA}
	Instance hashCount_ As Long
	Instance dataCount_ As Long
	Instance ubData_ As Long
	Instance dataResizeAmount_ As Long

	Class Method Create()
		dataResizeAmount_ = 100
	End Method

	Interface i{TKEY}To{TDATA}Map
		Inherit IUnknown

		Property Get HashCount() As Long
			Property = hashCount_
		End Property
		Property Set HashCount( value As Long )
			Me.ResizeHash( value )
		End Property

		Property Get ResizeAmount() As Long
			Property = dataResizeAmount_
		End Property
		Property Set ResizeAmount( ByVal value As Long )
			If value>0 Then dataResizeAmount_ = value
		End Property

		Property Get Count() As Long
			Property = dataCount_
		End Property
		Property Set Count( ByVal value As Long )
			If value> dataCount_ Then Me.ResizeData( value )
		End Property

		Method Clear()
			Me.ResizeHash( 0 )
			Me.ResizeData( 0 )
		End Method

		' Returns index of added item
		Method Add( {KEYREF} pKey As {TKEY}, {DATAREF} value As {TDATA} ) As Long
			Register index As Long
			Register lastIndex As Long
			Local hc As Long

			hc = {HASH}( pKey, hashCount_ )

			index = hash_( hc )

			If index = -1 Then
				' First entry
				If dataCount_>=ubData_-1 Then Me.ResizeData( ubData_+dataResizeAmount_ )
				hash_( hc ) = dataCount_
			Else
				' Collision
				Do
					If {KEYCMP}( pKey, key_( index ) ) Then
						lastIndex = index
						index = next_( index )
					Else
						Method = index
						Exit Method
					End If
				Loop While index>-1
				If dataCount_>=ubData_-1 Then Me.ResizeData( ubData_+dataResizeAmount_ )
				next_( lastIndex ) = dataCount_
			End If
			next_( dataCount_ ) = -1
			key_( dataCount_ ) = pKey
			value_( dataCount_ ) = value
			Method = dataCount_
			Incr dataCount_
		End Method

		Property Get Contains( {KEYREF} pKey As {TKEY} ) As Long
			' -1 if found, 0 if not found
			Property = ( Me.Index( pKey ) > -1 )
		End Property

		Property Get Index( {KEYREF} pKey As {TKEY} ) As Long
			' Returns index of key
			' Returns -1 if not found
			' Returns 0 - dataCount_-1 if found
			Register index As Long
			Local hc As Long

			hc = {HASH}( pKey, hashCount_ )
			index = hash_( hc )

			While index > -1
				If {KEYCMP}( pKey, key_( index ) ) Then
					index = next_( index )
				Else
					Property = index
					Exit Property
				End If
			Wend
			Property = -1
		End Property

		Property Get Key( ByVal index As Long ) As {TKEY}
			If index>=0 And index< dataCount_ Then
				Property = key_( index )
			End If
		End Property

		Property Get Value( ByVal index As Long ) As {TDATA}
			If index>=0 And index<dataCount_ Then
				Property = value_( index )
			End If
		End Property
		Property Set Value( ByVal index As Long, {DATAREF} sf As {TDATA} )
			If index>=0 And index<dataCount_ Then
				Reset value_( index )
				value_( index ) = sf
			End If
		End Property

		Property Get Item( {KEYREF} pKey As {TKEY} ) As {TDATA}
			Local index As Long
			index = Me.Index( pKey )
			If index>-1 Then Property = value_( index )
		End Property
		Property Set Item( {KEYREF} pKey As {TKEY}, {DATAREF} value As {TDATA} )
			Local index As Long
			index = Me.Index( pKey )
			If index>-1 Then
				value_( index ) = value
			End If
		End Property
	End Interface

	Class Method ResizeData( ByVal value As Long )
		If value>0 Then
			ReDim Preserve next_( value-1 )
			ReDim Preserve key_( value-1 )
			ReDim Preserve value_( value-1 )
			ubData_ = value
			dataCount_ = Min&( ubData_-1, dataCount_ )
		Else
			Erase next_()
			Erase key_()
			Erase value_()
			ubData_ = 0
			dataCount_ = 0
		End If
	End Method

	Class Method ResizeHash( ByVal value As Long )
		value = CollectionSize( value ): ' Make a prime >= value
		ReDim hash_( value-1 )
		Memory Fill VarPtr( hash_(0) ), value*SizeOf(value), Byte &HFF
		hashCount_ = value
	End Method
End Class
