' A Circular FIFO collection of {TVAL}s.
' FIFO - First in First out - Items are removed in the order they were added.
'
'Interface i{TVAL}Queue
'
'Property Count() As Long - Get the number of items on the Queue
'Property IsEmpty() as Long - True if empty, False if it has items
'Method Add( {VALREF} value As {TVAL} ) - Add an item to the Queue
'Method Remove() As {TVAL} - Remove an item from the Queue
'
'Method Clear() - Clear the Queue
'Property Capacity() As Long - Get/Set the Queue capacity

Sub New{TVAL}Queue( obj As i{TVAL}Queue, Optional capacity As Long, Optional resizeAmount As Long )
	obj = Class "c{TVAL}Queue"
	If Not IsMissing( capacity ) Then
		obj.Capacity = capacity
	End If
	If Not IsMissing( resizeAmount ) Then
		obj.ResizeAmount = resizeAmount
	End If
End Sub

Class c{TVAL}Queue
	Instance count_ As Long
	Instance head_ As Long: ' Where next item goes
	Instance tail_ As Long: ' Where next item to remove is
	Instance values_() As {TVAL}
	Instance resizeAmount_ As Long: ' How many items to add at a time

	Class Method Create()
		resizeAmount_ = 10
	End Method

	Interface i{TVAL}Queue
		Inherit IUnknown

		Property Get Count() As Long
			Property = IIf&( head_>tail_, head_-tail_, count_ + (head_-tail_) )
		End Property

		Property Get IsEmpty() As Long
			Property = (tail_=head_)
		End Property

		Method Add( {VALREF} value As {TVAL} )
			Local nxt As Long
			nxt = (head_ + 1) Mod count_
			If nxt = tail_ Then
				Me.Capacity = Me.Capacity + Me.ResizeAmount
			End If
			values_( head_ ) = value
			head_ = nxt
		End Method

		Method Remove() As {TVAL}
			If tail_ <> head_ Then
				Method = values_( tail_ )
				tail_ = (tail_ + 1) Mod count_
			End If
		End Method

		Method Clear()
			Local i As Long
			For i=0 To count_
				Reset values_( i )
			Next
			head_ = 0
			tail_ = 0
		End Method

		Property Get ResizeAmount() As Long
			Property = Min&( resizeAmount_, 1 )
		End Property
		Property Set ResizeAmount( ByVal value As Long )
			If value>0 Then resizeAmount_ = value
		End Property

		Property Get Capacity() As Long
			Property = count_
		End Property
		Property Set Capacity( ByVal value As Long )
			Local tmp() As {TVAL}
			Local i, c, h, t As Long

			If value>0 Then
				c =UBound( values_() )
				If c>0 Then
					' Save values
					ReDim tmp( c )
					For i=0 To c
						tmp(i) = values_(i)
					Next
				End If
				' Redim to new dimension
				ReDim values_( value-1 )
				h = head_
				t = tail_
				head_ = 0
				tail_ = 0
				count_ = value
				' Copy any existing values back
				While t<>h
					Me.Add( tmp(t) )
					t = (t+1) Mod count_
				Wend
			Else
				Erase values_()
				head_ = 0
				tail_ = 0
				count_ = 0
			End If
		End Property
	End Interface
End Class
